-------------------------------------------------------------------------------
-- Localized Lua globals.
-------------------------------------------------------------------------------
local _G = getfenv(0)

-- Functions
local pairs = _G.pairs


-- Libraries
local string = _G.string
local table = _G.table


-------------------------------------------------------------------------------
-- AddOn namespace.
-------------------------------------------------------------------------------
local FOLDER_NAME, private = ...

local LibStub = _G.LibStub
local addon = LibStub("AceAddon-3.0"):NewAddon(FOLDER_NAME, "AceEvent-3.0", "AceHook-3.0", "AceTimer-3.0")

private.pet_list = {}

local known_pets_by_npc_id = {}
local unknown_pets_by_npc_id = {}

local known_pets_by_name = {}
local unknown_pets_by_name = {}

local pets_by_npc_id = {}

local pet_registry = {}

-------------------------------------------------------------------------------
-- Constants.
-------------------------------------------------------------------------------
local MAX_ACTIVE_PETS = 3


-------------------------------------------------------------------------------
-- Helper functions.
-------------------------------------------------------------------------------
local AcquireTable
local ReleaseTable
do
	local table_cache = {}


	function AcquireTable()
		return table.remove(table_cache) or {}
	end


	function ReleaseTable(tbl)
		if not tbl then
			return
		end
		table.wipe(tbl)
		table.insert(table_cache, tbl)
	end
end -- do block


local ModifyTooltip
do
	local output = {}
	local temp = {}

	function ModifyTooltip(identifier, known_table, unknown_table)
		local known_entries = known_table[identifier]
		local unknown_entries = unknown_table[identifier]
		table.wipe(output)

		if known_entries then
			local shown_header

			for index = 1, #known_entries do
				local pet = known_entries[index]
				table.wipe(temp)

				if not shown_header then
					table.insert(temp, "\n")
					table.insert(temp, pet.name)
					table.insert(temp, _G.HEADER_COLON)

					if #known_entries < MAX_ACTIVE_PETS then
						table.insert(temp, "\n")
						table.insert(temp, "|cff82c5ff")
						table.insert(temp, _G.UNIT_CAPTURABLE)
						table.insert(temp, "|r")
					end
					table.insert(temp, "\n|cffffffff")
					table.insert(temp, _G.BATTLE_PET_CAPTURED)
					table.insert(temp, _G.HEADER_COLON)
					table.insert(temp, "|r\n")
					shown_header = true
				end
				table.insert(temp, _G.ITEM_QUALITY_COLORS[pet.rarity - 1].hex)
				table.insert(temp, "    ")
				table.insert(temp, _G["BATTLE_PET_BREED_QUALITY" .. pet.rarity])
				table.insert(temp, "|r |cffffffff")
				table.insert(temp, _G.PARENS_TEMPLATE:format(_G.TOOLTIP_UNIT_LEVEL:format(pet.level)))
				table.insert(temp, "|r")

				table.insert(output, table.concat(temp, ""))
			end
		elseif unknown_entries then
			local pet = unknown_entries[1]
			table.wipe(temp)

			table.insert(temp, "\n")
			table.insert(temp, pet.name)
			table.insert(temp, _G.HEADER_COLON)
			table.insert(temp, "\n")
			table.insert(temp, "|cff82c5ff")
			table.insert(temp, _G.UNIT_CAPTURABLE)
			table.insert(temp, "|r")

			table.insert(output, table.concat(temp, ""))
		end

		if #output > 0 then
			_G.GameTooltip:AddLine(table.concat(output, "\n"))
		end
	end
end


do
	local previous_tooltip_line
	local entity_registry = {}

	_G.GameTooltip:HookScript("OnUpdate", function(self)
		if not _G.Minimap:IsMouseOver() then
			return
		end
		local tooltip_line = _G["GameTooltipTextLeft1"]:GetText()

		if not tooltip_line or tooltip_line == previous_tooltip_line then
			return
		end
		previous_tooltip_line = tooltip_line
		tooltip_line = tooltip_line:gsub([[|T(.-)|t]], "")

		if tooltip_line:find("\n") then
			local entities = { ("\n"):split(tooltip_line) }
			table.wipe(entity_registry)

			for index = 1, #entities do
				local entity = entities[index]

				if not entity_registry[entity] then
					ModifyTooltip(entity, known_pets_by_name, unknown_pets_by_name)
					entity_registry[entity] = true
				end
			end
		else
			ModifyTooltip(tooltip_line, known_pets_by_name, unknown_pets_by_name)
		end
		_G.GameTooltip:Show()
	end)

	_G.GameTooltip:HookScript("OnHide", function()
		previous_tooltip_line = nil
	end)
end -- do-block


do
	local PET_TEXTURES = {
		"Border",
		"BorderAlive",
	}

	local PET_FONTS = {
		"Level",
		"Name",
	}

	_G.hooksecurefunc("PetBattleUnitFrame_UpdateDisplay", function(frame)
		if not frame.petOwner or not frame.petIndex then
			return
		end
		local r, g, b, hex = _G.GetItemQualityColor(_G.C_PetBattles.GetBreedQuality(frame.petOwner, frame.petIndex) - 1)

		for index = 1, #PET_TEXTURES do
			if frame[PET_TEXTURES[index]] then
				frame[PET_TEXTURES[index]]:SetVertexColor(r, g, b)
			end
		end

		for index = 1, #PET_FONTS do
			local font_string = frame[PET_FONTS[index]]

			if font_string then
				local text = font_string:GetText()

				if text then
					font_string:SetFormattedText("|c%s%s|r", hex, text)
				end
			end
		end
	end)
end -- do-block

function private.UpdatePetList()
	for npc_id, pet_list in pairs(pets_by_npc_id) do
		ReleaseTable(pet_list)
	end
	table.wipe(pets_by_npc_id)
	table.wipe(private.pet_list)

	local num_pets = _G.C_PetJournal.GetNumPets(_G.PetJournal.isWild)

	for pet_index = 1, num_pets do
		local pet_id, _, is_owned, _, level, _, _, name, _, pet_type, creature_id, _, _, _, can_battle = _G.C_PetJournal.GetPetInfoByIndex(pet_index, _G.PetJournal.isWild)
		local _, max_health, attack, speed, rarity = _G.C_PetJournal.GetPetStats(pet_id)

		local identifier = (pet_id ~= 0) and pet_id or name
		local pet = pet_registry[identifier]

		if not pet then
			pet = {
				can_battle = can_battle,
				id = pet_id,
				journal_index = pet_index,
				name = name,
				rarity = rarity,
				type = pet_type,
			}
			pet_registry[identifier] = pet
		end
		pet.attack = attack
		pet.is_owned = is_owned
		pet.level = level
		pet.max_health = max_health
		pet.speed = speed

		private.pet_list[#private.pet_list + 1] = pet

		pets_by_npc_id[creature_id] = pets_by_npc_id[creature_id] or AcquireTable()
		pets_by_npc_id[creature_id][#pets_by_npc_id[creature_id] + 1] = pet
	end
	table.wipe(known_pets_by_npc_id)
	table.wipe(unknown_pets_by_npc_id)

	table.wipe(known_pets_by_name)
	table.wipe(unknown_pets_by_name)

	for creature_id, pet_list in pairs(pets_by_npc_id) do
		if pet_list[1].is_owned then
			known_pets_by_npc_id[creature_id] = pet_list
			known_pets_by_name[pet_list[1].name] = pet_list
		else
			unknown_pets_by_npc_id[creature_id] = pet_list
			unknown_pets_by_name[pet_list[1].name] = pet_list
		end
	end
	-- TODO: Sort here.
	--	private.PetJournal_FindPetCardIndex()
	--	private.PetJournal_UpdatePetList()
	_G.PetJournal_FindPetCardIndex()
	_G.PetJournal_UpdatePetList()
end


-------------------------------------------------------------------------------
-- Initialization functions.
-------------------------------------------------------------------------------
function addon:OnInitialize()
	self:RegisterEvent("ADDON_LOADED")
end


function addon:OnEnable(event_name)
	-- Trickery to get the the initial pet list.
	self:ScheduleTimer(function()
		local sfx = _G.tonumber(_G.GetCVar("Sound_EnableSFX"))
		_G.SetCVar("Sound_EnableSFX", 0)
		_G.TogglePetJournal(2)
		_G.HideUIPanel(_G.PetJournalParent)
		_G.SetCVar("Sound_EnableSFX", sfx)
	end, 2)
end


-------------------------------------------------------------------------------
-- Event handlers.
-------------------------------------------------------------------------------
function addon:ADDON_LOADED(event_name, addon_name)
	if addon_name ~= "Blizzard_PetJournal" then
		return
	end
	self:HookScript(_G.PetJournalParent, "OnShow", private.UpdatePetList)
	--	self:RawHook("PetJournal_FindPetCardIndex", private.PetJournal_FindPetCardIndex, true)
	--	self:RawHook("PetJournal_UpdatePetList", private.PetJournal_UpdatePetList, true)

	self:UnregisterEvent("ADDON_LOADED")
	self:RegisterEvent("PET_JOURNAL_LIST_UPDATE")
	self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
end


function addon:PET_JOURNAL_LIST_UPDATE(event_name)
	private.UpdatePetList()
end


function addon:UPDATE_MOUSEOVER_UNIT()
	if not _G.UnitIsWildBattlePet("mouseover") then
		return
	end
	ModifyTooltip(_G.tonumber(_G.UnitGUID("mouseover"):sub(6, 10), 16), known_pets_by_npc_id, unknown_pets_by_npc_id)
	_G.GameTooltip:Show()
end
